/* Generic C includes */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

/* Structures, protos and erreurs of the lib */
#include "libregionMap.h"
 
#ifndef MAX
#define MAX(a,b) ((a) > (b) ? (a) : (b))
#endif
#ifndef MIN
#define MIN(a,b) ((a) < (b) ? (a) : (b))
#endif

/**********************************************************************
 *
 * Reads map made on-board Ressac
 *
 **********************************************************************/

int
rMap_readRessacMap (FILE*	file,
		REGION_MAP*	regionMap)
{
	RMAP_REGION*		region;
	int			i, j, k;

	double		north_northWestCorner, east_northWestCorner;
	double		scale;
	double		northernSize, easternSize;
	int		width, height;
	int*		ressacMap;
	int*		mapPtr;
	int		flat, unknown, obstacle;
	int		currentDate;
	int		iMap_NorthEast, jMap_NorthEast;
	int		iMap, jMap;
	int nbOut = 0;

	RMAP_FUSION_TYPE	fusionType = RMAP_FUSION_OVERRIDE;
	double			sumProbas;

	/* The following variables are for the Hack1 and Hack2 to palliate some
	   bugs (or erroneous behavior) of the EVA process */
	double		xShift = -3.0;
	double		yShift = 4.0;
	int		xShift_pixel, yShift_pixel;
	int		size = regionMap->robotCellRadius;
	int		l;

	/* Read the header of the map file */
	fread (&north_northWestCorner, sizeof(double), 1, file);
	fread (&east_northWestCorner, sizeof(double), 1, file);
	fread (&height, sizeof(int), 1, file);
	fread (&width, sizeof(int), 1, file);
	fread (&scale, sizeof(double), 1, file);

	printf ("Map scale: %.2f\n", scale);
	printf ("Northern extent of the map: %2.f meters, %d pixels\n", 
		height*scale, height);
	printf ("Eastern extent of the map: %2.f meters, %d pixels\n", 
		width*scale, width);
	printf ("north_northWestCorner: %.2f\n", north_northWestCorner);
	printf ("east_northWestCorner: %.2f\n", east_northWestCorner);

	if (fabs (scale - regionMap->scale) > 0.001) {
		rMapError (ERR_RMAP_SCALE);
		return (RMAP_ERROR);
	}

	iMap_NorthEast 
	  = ((east_northWestCorner - regionMap->origin2corner.euler.euler[3])
	     / regionMap->scale);
	jMap_NorthEast 
	  = ((north_northWestCorner - regionMap->origin2corner.euler.euler[4])
	     / regionMap->scale);

	xShift_pixel = (int) (xShift / scale);
	yShift_pixel = (int) (yShift / scale);

	/* Allocates the array of values */
	ressacMap = malloc (3 * height * width * sizeof (int));
	if (ressacMap == NULL) 
	  return rMapError(ERR_RMAP_ALLOC_REGION_MAP);
	/* Read the array of data */
	fread (ressacMap, sizeof(int), (3 * width * height), file);

	/* Initialize and the number of newly perceived cells */
	regionMap->nbNewCells = 0;

	/* Loop on the regionMap cells, and compute naive probabilities. */
	/* Note on the way we loop in the map:
	   - For a regionMap, i (resp. j) spans the X/East axis (resp. Y/North)
	   - For the ressacMap, lines (resp columns) span the east (resp. north)
	   axes.
	   We loop with (i,j) in the ressacMap convention (on the ressacMap
	   pointer, a "++" is a move towards East, a "+= width" is a move toward
	   South). 
	*/
	currentDate = regionMap->nbUpdates + 1;
	mapPtr = ressacMap;
	for (i = 0; i < width; i++) {
	  for (j = 0; j < height; j++) {
	    iMap = j + iMap_NorthEast;
	    jMap = - i + jMap_NorthEast;

	    /* Hack1: shift the regions coordinates */
	    iMap += xShift_pixel;
	    jMap += yShift_pixel;

	    flat = *mapPtr;
	    mapPtr++;
	    obstacle = *mapPtr;
	    mapPtr++;
	    unknown = *mapPtr;
	    mapPtr++;

	    /* Check that the considered point is within the map */
	    if ((iMap >= 0) && (iMap < regionMap->nblig)
		&& (jMap >= 0) && (jMap < regionMap->nbcol)) {
	      region = &(regionMap->regions[iMap * regionMap->nbcol + jMap]);
	      /* Do something only if the ressacMap pixel has been perceived
		 AND if the region has not been initally set */
	      if (((obstacle != 0) || (flat != 0)) 
		  && (region->updateState != RMAP_REGION_INITIALLY_PERCEIVED)) {
		/* This is a perceived region, update some global region and
		   regionMap characteristics  */
		if (region->lastDate != currentDate) {
		  rMap_updateStates (regionMap, region, iMap, jMap, currentDate);
		}
		/* Compute the probabilities with the information received from
		   ressac (quite dummy version) */
		region->classificationType[CLASSIFICATION_3D] = 1;
		region->infos3d.classificationState = EXTERNALLY_CLASSIFIED;
		region->infos3d.probaState = RMAP_PROBAS_COMPUTED;
		region->infos3d.newProbas[RMAP_OBSTACLE] 
		  = (double) obstacle / (obstacle + flat);
		region->infos3d.newProbas[RMAP_FLAT]
		  = (double) flat / (obstacle + flat);


		/* Hack2 : mark as RMAP_REGION_NOT_ENOUGH_POINTS the ones for
		   which the 'unknown' counter is the biggest */
		if (obstacle > flat) {
		  if (unknown >= obstacle) 
		    region->updateState = RMAP_REGION_NOT_ENOUGH_POINTS; 
		} else {
		  if (unknown >= flat) 
		    region->updateState = RMAP_REGION_NOT_ENOUGH_POINTS; 
		} /* End of Hack2 */


	      }
	    }
	    else {
	      nbOut++;
	    }
	  }
	}

	/* Hack2 : now loops on the updated regions, mark the ones for which
	   unknown is the biggest counter and that are close to unperceived
	   cells as unperceived */
	for (i = MAX (regionMap->last_iMin - size, size); 
	     i < MIN ((regionMap->nblig - size - 1), regionMap->last_iMax + size); 
	     i++) {
	  for (j = MAX (regionMap->last_jMin - size, size);
	       j < MIN ((regionMap->nbcol - size - 1), regionMap->last_jMax + size); 
	       j++) {
	    /* Are we on an unperceived region ? */
	    region = &(regionMap->regions[i * regionMap->nbcol + j]);
	    if (region->updateState == RMAP_REGION_NEVER_PERCEIVED) {
	      /* Yes we are: all the cells closer to size that have previously
		 been marked as RMAP_REGION_NOT_ENOUGH_POINTS are now marked as
		 never perceived */
	      for (k = i - size; k <= i + size; k++) {
		for (l = j - size; l <= j + size; l++) {
		  if (sqrt ((double) (k - i)*(k - i) + (l - j)*(l - j)) 
		      <= (double) (size)) {
		    region = &(regionMap->regions[k * regionMap->nbcol + l]);
		    if (region->updateState == RMAP_REGION_NOT_ENOUGH_POINTS) {
		      region->updateState = RMAP_REGION_NEWLY_PERCEIVED;
		      region->infos3d.newProbas[RMAP_OBSTACLE] = 0.5;
		      region->infos3d.newProbas[RMAP_FLAT] = 0.5;
		    }
		  }
		}
	      }
	    }
	  }
	}
	/* Now re-mark the remaining RMAP_REGION_NOT_ENOUGH_POINTS as newly
	   perceived */
	for (i = MAX (regionMap->last_iMin - size, size); 
	     i < MIN ((regionMap->nblig - size - 1), regionMap->last_iMax + size); 
	     i++) {
	  for (j = MAX (regionMap->last_jMin - size, size);
	       j < MIN ((regionMap->nbcol - size - 1), regionMap->last_jMax + size); 
	       j++) {
	    region = &(regionMap->regions[i * regionMap->nbcol + j]);
	    if (region->updateState == RMAP_REGION_NOT_ENOUGH_POINTS) 
	      region->updateState = RMAP_REGION_NEWLY_PERCEIVED;
	  }
	} /* End of Hack2 */

	
	printf ("Nb points of the ressacMap out of the regionMap: %d\n", nbOut);
	/* Update the various global informations` on the regionMap */
	rMap_endUpdate (regionMap);

	/* Merge the new probabilities with the existing ones */
	for (i = regionMap->last_iMin; i < regionMap->last_iMax; i++) {
	  for (j = regionMap->last_jMin; j < regionMap->last_jMax; j++) {
	    region = regionMap->regions + (regionMap->nbcol * i) + j;
	    if (region->updateState == RMAP_REGION_NEWLY_PERCEIVED) {
	      switch (fusionType) {
	      case RMAP_FUSION_OVERRIDE:
		memcpy (region->infos3d.mergedProbas, 
			region->infos3d.newProbas,
			NB_3D_CLASSES * sizeof(double));
		region->infos3d.probaState = RMAP_PROBAS_COMPUTED;
		break;
	      case RMAP_FUSION_MERGE:
		sumProbas = 0.0;
		for (k = 0; k < NB_3D_CLASSES; k++) {
		  sumProbas += (region->infos3d.newProbas[k] 
				* region->infos3d.mergedProbas[k]);
		}
		for (k = 0; k < NB_3D_CLASSES; k++) {
		  region->infos3d.mergedProbas[k] 
		    = (region->infos3d.newProbas[k] 
		       * region->infos3d.mergedProbas[k]) / sumProbas;
		}
		region->infos3d.probaState = RMAP_PROBAS_COMPUTED;
		break;
	      default:
		printf ("Can not handle fusion type %d\n", fusionType);
		return rMapError (ERR_RMAP_IMPOSSIBLE_CASE);
		/* 	    break; Unreached */
	      }
	    }
	  }
	}

	/* Grow the obstacle regions */
	rMap_computeGrowProbas (regionMap);
	
	free (ressacMap);
	return (RMAP_OK);
}

